import os
import re
import spacy
from spacy.matcher import Matcher
import re
import pandas as pdas
from nltk.corpus import stopwords
from os import path
from glob import glob
import nltk

# TODO , See if this need to be done in another place, performance improvements.
nltk.download("stopwords")
# TODO End
config_folder = "data"


def set_config_folder(new_location):
    global config_folder
    config_folder = new_location


from pyresumize.interfaces import (
    EducationBaseInterface,
    NameBaseInterface,
    PhoneBaseInterface,
    EmailBaseInterface,
    SkillBaseInterface,
    EmployerBaseInterface,
)

from pyresumize.utilities import Utilities


EDUCATION = [
    "BE",
    "B.E.",
    "B.E",
    "BS",
    "B.S",
    "ME",
    "M.E",
    "M.E.",
    "MS",
    "M.S",
    "BTECH",
    "B.TECH",
    "M.TECH",
    "MTECH",
    "SSC",
    "HSC",
    "CBSE",
    "ICSE",
    "X",
    "XII",
]
STOPWORDS = set(stopwords.words("english"))

"""
Below Modules are standard implementations of the respective interfaces
The developer of the library can extend the functionality as in needed by pluggin in a custom module
"""


class EducationStandardEngine(EducationBaseInterface):
    def __init__(self, nlp) -> None:
        super().__init__(config_folder)
        self.nlp = nlp
        pass

    def process(self, extracted_text):
        _text = self.nlp(extracted_text)
        _text = [sent.text.strip() for sent in _text.sents]

        degrees = {}

        # Lets search for Single words like BE
        for index, text in enumerate(_text):
            for tex in text.split():
                # Replace all special symbols
                tex = re.sub(r"[?|$|.|!|,]", r"", tex)
                if tex.upper() in EDUCATION and tex not in STOPWORDS:
                    if len(_text) > index + 1:
                        degrees[tex] = text + _text[index + 1]
        # Now Lets search Longer word like Bachelor of Technology
        # for the Combined works such as Operating Systems
        # TODO , Use Noun Chunks
        # Improve this please
        education = {}
        education_map = []
        for key in degrees.keys():
            year = re.search(
                re.compile(r"(((20)(\d{2})))"), degrees[key]
            )  # Assumes that the year is somewhere in 2000s
            if year:
                education_map.append((key, "".join(year[0])))
            else:
                education_map.append(key)
        education["entry"] = education_map
        education["universities"] = self.__find_universities(extracted_text)
        # Later can be added with university
        return education

    def __find_universities(self, resume_text):
        """ """
        nlp_text = self.nlp(resume_text)

        tokens = [token.text for token in nlp_text if not token.is_stop]
        universities_input_folder = os.path.join(self.config_folder, "universities")
        utils = Utilities()
        universities = utils.generate_keywords_from_csv_files(universities_input_folder)
        universities = list(map(lambda x: str(x).lower(), universities))  # Normalising the Strings to Lower

        candidate_universities = []

        for token in nlp_text.noun_chunks:
            token = token.text.lower().strip()  # for the Combined works such as Operating Systems
            if token in universities:
                candidate_universities.append(token)

        return candidate_universities


class PhoneStandardEngine(PhoneBaseInterface):
    def __init__(self, nlp) -> None:
        super().__init__(config_folder)
        self.nlp = nlp

    def process(self, extracted_text):
        results = set(re.findall(r"[\+\(]?[1-9][0-9 .\-\(\)]{8,}[0-9]", extracted_text))
        phone = ",".join(results)
        return phone


class NameStandardEngine(NameBaseInterface):
    def __init__(self, nlp) -> None:
        super().__init__(config_folder)
        self.matcher = Matcher(nlp.vocab)
        self.nlp = nlp

    def process(self, extracted_text):
        nlp_text = self.nlp(extracted_text)
        # Identify the Names (Nouns)
        pattern = [{"POS": "PROPN"}, {"POS": "PROPN"}]

        self.matcher.add("NAME", [pattern])

        matches = self.matcher(nlp_text)

        for match_id, start, end in matches:
            span = nlp_text[start:end]
            return span.text
        return ""


class EmployerStandardEngine(EmployerBaseInterface):
    def __init__(self, nlp) -> None:
        super().__init__(config_folder)
        self.nlp = nlp
        pass

    def process(self, extracted_text):
        """does nothing"""
        candidate_employment = []
        nlp_text = self.nlp(extracted_text)

        tokens = [token.text for token in nlp_text if not token.is_stop]
        employ_input_folder = os.path.join(self.config_folder, "employers")
        utils = Utilities()
        employers = utils.generate_keywords_from_csv_files(employ_input_folder)
        employers = list(map(lambda x: str(x).lower(), employers))  # Normalising the Strings to Lower

        candidate_employment = []

        for token in nlp_text.noun_chunks:
            token = token.text.lower().strip()  # for the Combined works such as Operating Systems
            if token in employers:
                if token not in candidate_employment:
                    candidate_employment.append(token)
        return candidate_employment


class SkillStandardEngine(SkillBaseInterface):
    def __init__(self, nlp) -> None:
        super().__init__(config_folder)
        self.nlp = nlp

    def process(self, extracted_text):
        """ """
        nlp_text = self.nlp(extracted_text)
        skills = []
        tokens = [token.text for token in nlp_text if not token.is_stop]
        # Iterate through the skills csvs and build info

        skills_input_folder = os.path.join(self.config_folder, "skills")
        utils = Utilities()
        skills = utils.generate_keywords_from_csv_files(skills_input_folder)
        skills = list(map(lambda x: str(x).lower(), skills))  # Normalising the Strings to Lower

        skillset = []

        # For words like JAVA
        for token in tokens:
            if token.lower() in skills:
                skillset.append(token)

        # for the Combined works such as Operating Systems
        for token in nlp_text.noun_chunks:
            token = token.text.lower().strip()
            if token in skills:
                skillset.append(token)
        # Lets mark it lower
        skillset = list(map(lambda x: str(x).lower(), skillset))
        # Remove duplicate skills, if any
        skillset = set(skillset)

        return skillset


class EmailStandardEngine(EmailBaseInterface):
    def __init__(self, nlp) -> None:
        super().__init__(config_folder)
        self.nlp = nlp
        pass

    def process(self, extracted_text):
        result = ""
        # Regex for finding email
        email = re.findall("([^@|\s]+@[^@]+\.[^@|\s]+)", extracted_text)
        if email:
            try:
                result = email[0].split()[0].strip(";")
                if result is None:
                    result = ""
            except IndexError:
                return result
        return result



class EmployerBaseInterface:
    """Idenify the previous employers"""

    def __init__(self, config_folder) -> None:
        self.config_folder = config_folder
        pass

    def process(self, extracted_text):
        """does nothing"""
        return []


class EducationBaseInterface:
    """Interface to fetch the education , graduation year and university"""

    def __init__(self, config_folder) -> None:
        self.config_folder = config_folder
        pass

    def process(self, extracted_text):
        """does nothing"""
        return {}


class EmailBaseInterface:
    """Interface to parse the Email address"""

    def __init__(self, config_folder) -> None:
        self.config_folder = config_folder
        pass

    def process(self, extracted_text):
        """does nothing"""
        return ""


class PhoneBaseInterface:
    """Interface to parse Phone number from resume"""

    def __init__(self, config_folder) -> None:
        self.config_folder = config_folder
        pass

    def process(self, extracted_text):
        """does nothing"""
        return ""


class NameBaseInterface:
    """Interface to parse the name in the resume"""

    def __init__(self, config_folder) -> None:
        self.config_folder = config_folder
        pass

    def process(self, extracted_text):
        """does nothing"""
        return ""


class SkillBaseInterface:
    """Interface to parse the Skills in a resume"""

    def __init__(self, config_folder) -> None:
        self.config_folder = config_folder
        pass

    def process(self, extracted_text):
        """does nothing"""
        return [""]



from os import path
from glob import glob
import pandas as pdas


class Utilities:
    """
    Like in any other module , we also have a utility
    """

    def generate_keywords_from_csv_files(self, foldername, csv_column=0):
        """Given CSV will be loaded and returned as a python list"""
        keywords = []
        files = glob(path.join(foldername, "*.csv"))
        for file in files:
            df = pdas.read_csv(file, on_bad_lines="skip", skiprows=1)
            ## Assumes that the first column is full of skills :)
            matrix2 = df[df.columns[csv_column]].to_numpy()
            new_kws = matrix2.tolist()
            keywords = keywords + new_kws
        return keywords

    def error_handler(self, error_string):
        """The guy who handles errors and inform the user"""
        data = {}
        # The Below Error Code Can be handled globally later :)
        data["error_code"] = -1
        data["error_string"] = error_string
        return data



from pdfminer.converter import TextConverter
from pdfminer.pdfinterp import PDFPageInterpreter
from pdfminer.pdfinterp import PDFResourceManager
from pdfminer.layout import LAParams
from pdfminer.pdfpage import PDFPage
import io
import spacy
from spacy.matcher import Matcher
import os
from pyresumize.utilities import Utilities
import pyresumize.modules as modules
from pyresumize.modules import (
    SkillStandardEngine,
    NameStandardEngine,
    PhoneStandardEngine,
    EmailStandardEngine,
    EducationStandardEngine,
    EmployerStandardEngine,
)


class Candidate:
    """The Person Class who owns the given resume
    The Extracted data is encapsulated in this class
    """

    name = ""
    phone = ""
    email = ""
    personal_details = {}
    skills = []
    education = {}
    employers = []


class ResumeEngine:
    """The GodFather Class with the API interface"""

    def __init__(self) -> None:
        self.nlp = spacy.load("en_core_web_sm")
        self.candidate = Candidate()
        self.name_engine = NameStandardEngine(self.nlp)
        self.skills_engine = SkillStandardEngine(self.nlp)
        self.phone_engine = PhoneStandardEngine(self.nlp)
        self.email_engine = EmailStandardEngine(self.nlp)
        self.education_engine = EducationStandardEngine(self.nlp)
        self.employer_engine = EmployerStandardEngine(self.nlp)

    def set_skills_engine(self, engine):
        self.skills_engine = engine

    def set_name_engine(self, engine):
        self.name_engine = engine

    def set_name_engine(self, engine):
        self.phone_engine = engine

    def set_email_engine(self, engine):
        self.email_engine = engine

    def set_education_engine(self, engine):
        self.education_engine = engine

    def set_employer_engine(self, engine):
        self.employer_engine = engine

    def set_custom_keywords_folder(self, folder_name):
        modules.set_config_folder(folder_name)

    def __generate_json(self):
        """
        Generates the respective json format of the resume
        """
        json_data = {}
        json_data["basic_details"] = self.candidate.personal_details
        json_data["skills"] = self.candidate.skills
        json_data["education"] = self.candidate.education
        json_data["employers"] = self.candidate.employers
        return json_data

    def check_file_sanity(self, file_path):
        """
        Utility function to check if the file given is valid
        """
        return self.__extract_text_from_pdf(file_path)

    def process_resume(self, file_path):
        """
        The Worker API !
        """
        resume_data = self.__extract_text_from_pdf(file_path)
        if resume_data is None:
            util = Utilities()
            return util.error_handler("File %s Can not be opened" % (file_path))
        self.candidate.name = self.name_engine.process(resume_data)
        self.candidate.skills = self.skills_engine.process(resume_data)
        self.candidate.personal_details["phone"] = self.phone_engine.process(resume_data)
        self.candidate.personal_details["email"] = self.email_engine.process(resume_data)
        self.candidate.education = self.education_engine.process(resume_data)
        self.candidate.employers=self.employer_engine.process(resume_data)

        data = self.__generate_json()
        # Check if file is a pdf / doc and process accordingly.
        return data

    def __extract_text_from_pdf(self, pdf_path):
        """Get All text from PDF"""
        text = ""
        try:
            with open(pdf_path, "rb") as fh:
                # iterate over all pages of PDF document
                for page in PDFPage.get_pages(fh, caching=True, check_extractable=True):
                    resource_manager = PDFResourceManager()
                    fake_file_handle = io.StringIO()
                    converter = TextConverter(resource_manager, fake_file_handle, codec="utf-8", laparams=LAParams())
                    page_interpreter = PDFPageInterpreter(resource_manager, converter)
                    page_interpreter.process_page(page)
                    text += fake_file_handle.getvalue()
                    fake_file_handle.close()
        except:
            # print("Unable to Process file %s"%pdf_path)
            # Any parsing failure we handle here .
            return None

        return text
